#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@author  : Hu Ji
@file    : container.py 
@time    : 2024/3/21
@site    : https://github.com/ice-black-tea
@software: PyCharm 

              ,----------------,              ,---------,
         ,-----------------------,          ,"        ,"|
       ,"                      ,"|        ,"        ,"  |
      +-----------------------+  |      ,"        ,"    |
      |  .-----------------.  |  |     +---------+      |
      |  |                 |  |  |     | -==----'|      |
      |  | $ sudo rm -rf / |  |  |     |         |      |
      |  |                 |  |  |/----|`---=    |      |
      |  |                 |  |  |   ,/|==== ooo |      ;
      |  |                 |  |  |  // |(((( [33]|    ,"
      |  `-----------------'  |," .;'| |((((     |  ,"
      +-----------------------+  ;;  | |         |,"
         /_)______________(_/  //'   | +---------+
    ___________________________/___  `,
   /  oooooooooooooooo  .o.  oooo /,   `,"-----------
  / ==ooooooooooooooo==.o.  ooo= //   ,``--{)B     ,"
 /_==__==========__==_ooo__ooo=_/'   /___________,"
"""
import contextlib
import sys
from argparse import Namespace, ArgumentParser
from subprocess import SubprocessError
from typing import Optional, List, Type, Dict, Tuple, Any

import yaml
from git import GitCommandError
from linktools import environ, ConfigError, utils
from linktools.cli import BaseCommand, subcommand, SubCommandWrapper, subcommand_argument, SubCommandGroup, \
    BaseCommandGroup, SubCommand
from linktools.cli.argparse import KeyValueAction, BooleanOptionalAction, ParserCompleter
from linktools.rich import confirm, choose

from .container import ContainerError, BaseContainer
from .manager import ContainerManager

manager = ContainerManager(environ)


def _iter_container_names():
    return [container.name for container in manager.containers.values()]


def _iter_installed_container_names():
    return [container.name for container in manager.get_installed_containers()]


class RepoCommand(BaseCommandGroup):
    """
    manage container repository
    """

    @property
    def name(self):
        return "repo"

    @subcommand("list", help="list repositories")
    def on_command_list(self):
        repos = manager.get_all_repos()
        for key, value in repos.items():
            data = {key: value}
            self.logger.info(
                yaml.dump(data, sort_keys=False).strip()
            )

    @subcommand("add", help="add repository")
    @subcommand_argument("url", help="repository url")
    @subcommand_argument("-b", "--branch", help="branch name")
    @subcommand_argument("-f", "--force", help="force add")
    def on_command_add(self, url: str, branch: str = None, force: bool = False):
        manager.add_repo(url, branch=branch, force=force)

    @subcommand("update", help="update repositories")
    @subcommand_argument("-f", "--force", help="force update")
    def on_command_update(self, force: bool = False):
        manager.update_repos(force=force)

    @subcommand("remove", help="remove repository")
    @subcommand_argument("url", nargs="?", help="repository url")
    def on_command_remove(self, url: str = None):
        repos = list(manager.get_all_repos().keys())
        if not repos:
            raise ContainerError("No repository found")

        if url is None:
            index = choose("Choose repository you want to remove", repos)
            if not confirm(f"Remove repository `{repos[index]}`?", default=False):
                raise ContainerError("Canceled")
            manager.remove_repo(repos[index])

        elif url in repos:
            if not confirm(f"Remove repository `{url}`?", default=False):
                raise ContainerError("Canceled")
            manager.remove_repo(url)

        else:
            raise ContainerError(f"Repository `{url}` not found.")


class ConfigCommand(BaseCommand):
    """
    manage container configs
    """

    @property
    def name(self):
        return "config"

    def init_arguments(self, parser: ArgumentParser) -> None:
        self.add_subcommands(parser)

    def run(self, args: Namespace) -> Optional[int]:
        subcommand = self.parse_subcommand(args)
        if subcommand:
            return subcommand.run(args)
        containers = manager.prepare_installed_containers()
        manager.create_docker_compose_process(
            containers,
            "config",
            privilege=False,
        ).check_call()

    @subcommand("set", help="set container configs")
    @subcommand_argument("configs", action=KeyValueAction, nargs="+", help="container config key=value")
    def on_command_set(self, configs: Dict[str, str]):
        manager.config.cache.save(**configs)
        for key in sorted(configs.keys()):
            value = manager.config.get(key)
            self.logger.info(f"{key}: {value}")

    @subcommand("unset", help="remove container configs")
    @subcommand_argument("configs", action=KeyValueAction, metavar="KEY", nargs="+", help="container config keys")
    def on_command_remove(self, configs: Dict[str, str]):
        manager.config.cache.remove(*configs)
        self.logger.info(f"Unset {', '.join(configs.keys())} success")

    @subcommand("list", help="list container configs")
    def on_command_list(self):
        keys = set()
        for container in manager.prepare_installed_containers():
            keys.update(container.configs.keys())
            if hasattr(container, "keys") and isinstance(container.keys, (Tuple, List, Dict)):
                keys.update([key for key in container.keys if key in manager.config])
        keys.update(manager.config.cache.keys())
        for key in sorted(keys):
            value = manager.config.get(key)
            self.logger.info(f"{key}: {value}")

    @subcommand("edit", help="edit the config file in an editor")
    @subcommand_argument("--editor", help="editor to use to edit the file")
    def on_command_edit(self, editor: str):
        return manager.create_process(editor, manager.config.cache.path).call()

    @subcommand("reload", help="reload container configs")
    def on_command_reload(self):
        manager.config.reload = True
        manager.prepare_installed_containers()


class ExecCommand(BaseCommand):
    """
    exec container command
    """

    @property
    def name(self):
        return "exec"

    @property
    def _subparser(self) -> ArgumentParser:
        parser = ArgumentParser()

        subcommands: List[SubCommand] = []
        for container in manager.get_installed_containers():
            subcommand_group = SubCommandGroup(container.name, container.description)
            subcommands.append(subcommand_group)
            subcommands.extend(self.walk_subcommands(container, parent_id=subcommand_group.id))
        self.add_subcommands(parser, target=subcommands)

        return parser

    def init_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("exec_name", nargs="?", metavar="CONTAINER", help="container name",
                            choices=utils.lazy_iter(_iter_installed_container_names))
        action = parser.add_argument("exec_args", nargs="...", metavar="ARGS", help="container exec args")

        if ParserCompleter:
            class Completer(ParserCompleter):
                get_parser = lambda _: self._subparser
                get_args = lambda _, args, **kw: [args.exec_name, *args.exec_args] if args.exec_name else None

            action.completer = Completer()

    def run(self, args: Namespace) -> Optional[int]:
        args = self._subparser.parse_args([args.exec_name, *args.exec_args] if args.exec_name else [])
        subcommand = self.parse_subcommand(args)
        if not subcommand or isinstance(subcommand, SubCommandGroup):
            return self.print_subcommands(args, root=subcommand, max_level=2)
        manager.prepare_installed_containers()
        return subcommand.run(args)


class Command(BaseCommandGroup):
    """
    Deploy and manage Docker/Podman containers with ease
    """

    @property
    def name(self) -> str:
        return "cntr"

    @property
    def parent(self) -> Optional[str]:
        return "common"

    @property
    def known_errors(self) -> List[Type[BaseException]]:
        return super().known_errors + [
            ContainerError, ConfigError, SubprocessError, GitCommandError, OSError
        ]

    def init_subcommands(self) -> Any:
        return [
            self,
            SubCommandWrapper(ExecCommand()),
            SubCommandWrapper(ConfigCommand()),
            SubCommandWrapper(RepoCommand()),
        ]

    @subcommand("list", help="list all containers")
    def on_command_list(self):
        install_containers = manager.get_installed_containers(resolve=False)
        all_install_containers = manager.resolve_depend_containers(install_containers)
        for container in sorted(manager.containers.values(), key=lambda o: o.order):
            if container not in all_install_containers:
                self.logger.info(f"[ ] {container.name}", extra={"style": "dim"})
            elif container in install_containers:
                self.logger.info(f"[*] {container.name} [added]", extra={"style": "red bold"})
            else:
                self.logger.info(f"[-] {container.name} [dependency]", extra={"style": "red dim"})

    @subcommand("add", help="add containers to installed list")
    @subcommand_argument("names", metavar="CONTAINER", nargs="+", help="container name",
                         choices=utils.lazy_iter(_iter_container_names))
    def on_command_add(self, names: List[str]):
        containers = manager.add_installed_containers(*names)
        if not containers:
            raise ContainerError("No container added")
        result = sorted(list([container.name for container in containers]))
        self.logger.info(f"Add {', '.join(result)} success")

    @subcommand("remove", help="remove containers from installed list")
    @subcommand_argument("-f", "--force", help="Force remove")
    @subcommand_argument("names", metavar="CONTAINER", nargs="+", help="container name",
                         choices=utils.lazy_iter(_iter_container_names))
    def on_command_remove(self, names: List[str], force: bool = False):
        containers = manager.remove_installed_containers(*names, force=force)
        if not containers:
            raise ContainerError("No container removed")
        result = sorted(list([container.name for container in containers]))
        self.logger.info(f"Remove {', '.join(result)} success")

    @subcommand("info", help="display container info")
    @subcommand_argument("names", metavar="CONTAINER", nargs="+", help="container name",
                         choices=utils.lazy_iter(_iter_container_names))
    def on_command_info(self, names: List[str]):
        for name in names:
            container = manager.containers[name]
            data = {
                name: {
                    "path": container.root_path,
                    "order": container.order,
                    "enable": container.enable,
                    "dependencies": container.dependencies,
                    "configs": list(set(container.configs.keys())),
                    "exposes": list(set([o.name for o in container.exposes])),
                }
            }
            self.logger.info(yaml.dump(data, sort_keys=False).strip())

    @subcommand("up", help="deploy installed containers")
    @subcommand_argument("--build", action=BooleanOptionalAction, help="build images before starting")
    @subcommand_argument("--pull", action=BooleanOptionalAction, help="always attempt to pull a newer version of the image")
    @subcommand_argument("name", metavar="CONTAINER", nargs="?", help="container name",
                         choices=utils.lazy_iter(_iter_installed_container_names))
    def on_command_up(self, name: str = None, build: bool = True, pull: str = False):
        containers = manager.prepare_installed_containers()
        target_containers = [c for c in containers if c.name == name] if name else containers

        build_options = []
        up_options = ["--detach", "--no-build"]
        if pull:
            build_options.extend(["--pull"])
            up_options.extend(["--pull", "always"])
        if not name:
            up_options.extend(["--remove-orphans"])

        services = []
        if name:
            services.extend(manager.containers[name].services.keys())
            if not services:
                raise ContainerError(f"No service found in container `{name}`")

        with self._notify_start(target_containers):
            if build:
                manager.create_docker_compose_process(
                    containers,
                    "build", *build_options, *services,
                ).check_call()
            manager.create_docker_compose_process(
                containers,
                "up", *up_options, *services
            ).check_call()

    @subcommand("restart", help="restart installed containers")
    @subcommand_argument("--build", action=BooleanOptionalAction, help="build images before starting")
    @subcommand_argument("--pull", action=BooleanOptionalAction, help="always attempt to pull a newer version of the image")
    @subcommand_argument("name", metavar="CONTAINER", nargs="?", help="container name",
                         choices=utils.lazy_iter(_iter_installed_container_names))
    def on_command_restart(self, name: str = None, build: bool = True, pull: str = False):
        containers = manager.prepare_installed_containers()
        target_containers = [c for c in containers if c.name == name] if name else containers

        build_options = []
        up_options = ["--detach", "--no-build"]
        if pull:
            build_options.extend(["--pull"])
            up_options.extend(["--pull", "always"])
        if not name:
            up_options.extend(["--remove-orphans"])

        services = []
        if name:
            services.extend(manager.containers[name].services.keys())
            if not services:
                raise ContainerError(f"No service found in container `{name}`")

        with self._notify_stop(target_containers):
            manager.create_docker_compose_process(
                containers,
                "stop", *services
            ).check_call()

        with self._notify_start(target_containers):
            if build:
                manager.create_docker_compose_process(
                    containers,
                    "build", *build_options, *services,
                ).check_call()
            manager.create_docker_compose_process(
                containers,
                "up", *up_options, *services
            ).check_call()

    @subcommand("down", help="stop installed containers")
    @subcommand_argument("name", metavar="CONTAINER", nargs="?", help="container name",
                         choices=utils.lazy_iter(_iter_installed_container_names))
    def on_command_down(self, name: str = None):
        containers = manager.prepare_installed_containers()
        target_containers = [c for c in containers if c.name == name] if name else containers

        services = []
        if name:
            services.extend(manager.containers[name].services.keys())
            if not services:
                raise ContainerError(f"No service found in container `{name}`")

        with self._notify_stop(target_containers):
            manager.create_docker_compose_process(
                containers,
                "down", *services
            ).check_call()

        with self._notify_remove(target_containers):
            pass

    @subcommand("update", help="update scripts and containers")
    def on_command_update(self):
        utils.create_process(
            sys.executable, "-m", "pip", "install", "-U", "linktools", "linktools-cntr"
        ).check_call()
        manager.update_repos(force=False)

    @classmethod
    @contextlib.contextmanager
    def _notify_start(cls, containers: List[BaseContainer]):
        for container in containers:
            if container.start_hooks:
                for hook in container.start_hooks:
                    hook()
            container.on_starting()

        yield

        for container in reversed(containers):
            container.on_started()

    @classmethod
    @contextlib.contextmanager
    def _notify_stop(cls, containers: List[BaseContainer]):
        for container in reversed(containers):
            container.on_stopping()

        yield

        for container in containers:
            container.on_stopped()
            if container.stop_hooks:
                for hook in container.stop_hooks:
                    hook()

    @classmethod
    @contextlib.contextmanager
    def _notify_remove(cls, containers: List[BaseContainer]):
        yield

        for container in containers:
            container.on_removed()


command = Command()
if __name__ == '__main__':
    command.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@author  : Hu Ji
@file    : deploy.py 
@time    : 2023/05/21
@site    :  
@software: PyCharm 

              ,----------------,              ,---------,
         ,-----------------------,          ,"        ,"|
       ,"                      ,"|        ,"        ,"  |
      +-----------------------+  |      ,"        ,"    |
      |  .-----------------.  |  |     +---------+      |
      |  |                 |  |  |     | -==----'|      |
      |  | $ sudo rm -rf / |  |  |     |         |      |
      |  |                 |  |  |/----|`---=    |      |
      |  |                 |  |  |   ,/|==== ooo |      ;
      |  |                 |  |  |  // |(((( [33]|    ,"
      |  `-----------------'  |," .;'| |((((     |  ,"
      +-----------------------+  ;;  | |         |,"
         /_)______________(_/  //'   | +---------+
    ___________________________/___  `,
   /  oooooooooooooooo  .o.  oooo /,   `,"-----------
  / ==ooooooooooooooo==.o.  ooo= //   ,``--{)B     ,"
 /_==__==========__==_ooo__ooo=_/'   /___________,"
"""

import yaml
from linktools import Config, utils
from linktools.decorator import cached_property

from linktools_cntr import BaseContainer
from linktools_cntr.container import ExposeMixin, ExposeLink, ExposeCategory


class Container(BaseContainer):

    @cached_property
    def configs(self):
        return dict(
            WILDCARD_DOMAIN=True,
            FLARE_TAG="latest",
            FLARE_DOAMIN=self.get_nginx_domain(""),
            FLARE_EXPOSE_PORT=Config.Alias(type=int, default=0),
            FLARE_ENABLE_LOGIN=Config.Confirm(default=False, cached=True),
            FLARE_USER=Config.Lazy(
                lambda cfg:
                Config.Prompt(default="admin", cached=True)
                if cfg.get("FLARE_ENABLE_LOGIN", type=bool)
                else ""
            ),
            FLARE_PASSWORD=Config.Lazy(
                lambda cfg:
                Config.Prompt(cached=True)
                if cfg.get("FLARE_ENABLE_LOGIN", type=bool)
                else ""
            )
        )

    @cached_property
    def exposes(self) -> [ExposeLink]:
        return [
            self.expose_other("在线工具集合", "tools", "", "https://tool.lu/"),
            self.expose_other("在线正则表达式", "regex", "", "https://regex101.com/"),
            self.expose_other("正则表达式手册", "regex", "", "https://tool.oschina.net/uploads/apidocs/jquery/regexp.html"),
            self.expose_other("在线json解析", "codeJson", "", "https://www.json.cn/"),
            self.expose_other("DNS查询", "dns", "", "https://tool.chinaz.com/dns/"),
            self.expose_other("图标下载", "progressDownload", "", "https://materialdesignicons.com/"),
        ]

    def on_starting(self):

        categories = {}
        apps = []
        bookmarks = []

        for key, value in vars(ExposeMixin).items():
            if isinstance(value, ExposeCategory):
                categories.setdefault(value, list())

        for container in sorted(self.manager.containers.values(), key=lambda o: o.order):
            for expose in container.exposes:
                if isinstance(expose, ExposeLink) and expose.is_valid:
                    categories[expose.category].append(expose)
                    if expose.category is self.expose_public:
                        apps.append(expose)
                    bookmarks.append(expose)

        data = {"links": []}
        for app in apps:
            data["links"].append({
                "name": app.name,
                "desc": app.desc,
                "icon": app.icon,
                "link": app.url,
            })
        utils.write_file(
            self.get_app_path("app", "apps.yml", create_parent=True),
            yaml.dump(data),
        )

        data = {"categories": [], "links": []}
        for category, links in categories.items():
            if not links:
                continue
            data["categories"].append({
                "id": category.name,
                "title": category.desc,
            })
            for link in links:
                data["links"].append({
                    "category": category.name,
                    "name": link.name,
                    "icon": link.icon,
                    "link": link.url,
                })
        utils.write_file(
            self.get_app_path("app", "bookmarks.yml", create_parent=True),
            yaml.dump(data),
        )

        self.write_nginx_conf(
            self.get_config("FLARE_DOAMIN"),
            self.get_path("nginx.conf"),
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@author  : Hu Ji
@file    : deploy.py 
@time    : 2023/05/21
@site    :  
@software: PyCharm 

              ,----------------,              ,---------,
         ,-----------------------,          ,"        ,"|
       ,"                      ,"|        ,"        ,"  |
      +-----------------------+  |      ,"        ,"    |
      |  .-----------------.  |  |     +---------+      |
      |  |                 |  |  |     | -==----'|      |
      |  | $ sudo rm -rf / |  |  |     |         |      |
      |  |                 |  |  |/----|`---=    |      |
      |  |                 |  |  |   ,/|==== ooo |      ;
      |  |                 |  |  |  // |(((( [33]|    ,"
      |  `-----------------'  |," .;'| |((((     |  ,"
      +-----------------------+  ;;  | |         |,"
         /_)______________(_/  //'   | +---------+
    ___________________________/___  `,
   /  oooooooooooooooo  .o.  oooo /,   `,"-----------
  / ==ooooooooooooooo==.o.  ooo= //   ,``--{)B     ,"
 /_==__==========__==_ooo__ooo=_/'   /___________,"
"""
from linktools import Config
from linktools.decorator import cached_property

from linktools_cntr import BaseContainer, ExposeLink


class Container(BaseContainer):

    @cached_property
    def configs(self):
        return dict(
            PORTAINER_TAG="alpine",
            PORTAINER_DOMAIN=self.get_nginx_domain(),
            PORTAINER_EXPOSE_PORT=Config.Alias(type=int, default=0),
        )

    @cached_property
    def exposes(self) -> [ExposeLink]:
        return [
            self.expose_public("Portainer", "docker", "Docker管理工具", self.load_nginx_url("PORTAINER_DOMAIN")),
            self.expose_container("Portainer", "docker", "Docker管理工具", self.load_port_url("PORTAINER_EXPOSE_PORT", https=False)),
        ]

    def on_starting(self):
        self.write_nginx_conf(
            self.get_config("PORTAINER_DOMAIN"),
            self.get_path("nginx.conf"),
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@author  : Hu Ji
@file    : deploy.py 
@time    : 2023/05/21
@site    :  
@software: PyCharm 

              ,----------------,              ,---------,
         ,-----------------------,          ,"        ,"|
       ,"                      ,"|        ,"        ,"  |
      +-----------------------+  |      ,"        ,"    |
      |  .-----------------.  |  |     +---------+      |
      |  |                 |  |  |     | -==----'|      |
      |  | $ sudo rm -rf / |  |  |     |         |      |
      |  |                 |  |  |/----|`---=    |      |
      |  |                 |  |  |   ,/|==== ooo |      ;
      |  |                 |  |  |  // |(((( [33]|    ,"
      |  `-----------------'  |," .;'| |((((     |  ,"
      +-----------------------+  ;;  | |         |,"
         /_)______________(_/  //'   | +---------+
    ___________________________/___  `,
   /  oooooooooooooooo  .o.  oooo /,   `,"-----------
  / ==ooooooooooooooo==.o.  ooo= //   ,``--{)B     ,"
 /_==__==========__==_ooo__ooo=_/'   /___________,"
"""
import os
import re
import shutil
import textwrap

from linktools import Config, utils
from linktools.decorator import cached_property

from linktools_cntr import BaseContainer


class Container(BaseContainer):

    @cached_property
    def keys(self):
        # dnsapi.txt 内容从 https://github.com/acmesh-official/acme.sh/wiki/dnsapi 拷贝
        path = os.path.join(os.path.dirname(__file__), "dnsapi.txt")
        data = utils.read_file(path, text=True)
        pattern = re.compile(r'export +(\w+)="?')
        return sorted(list(set(pattern.findall(data))))

    @cached_property
    def configs(self):
        return dict(
            NGINX_TAG="1.27.0-alpine",
            WILDCARD_DOMAIN=False,
            ROOT_DOMAIN=Config.Prompt(default="_", cached=True),
            HTTP_PORT=Config.Prompt(default=80, type=int, cached=True),
            HTTPS_ENABLE=Config.Confirm(default=True, cached=True),
            HTTPS_PORT=Config.Lazy(
                lambda cfg:
                Config.Prompt(type=int, default=443, cached=True)
                if cfg.get("HTTPS_ENABLE", type=bool)
                else Config.Alias(type=int, default=0)
            ),
            ACME_DNS_API=Config.Lazy(
                lambda cfg:
                Config.Error(textwrap.dedent(
                    """
                    Ensure ACME_DNS_API config matches --dns parameter in acme command is set.
                    · Also, set corresponding environment variables.
                    · For details, see: https://github.com/acmesh-official/acme.sh/wiki/dnsapi.
                    · Example command:
                      $ ct-cntr config set ACME_DNS_API=dns_ali Ali_Key=xxx Ali_Secret=yyy
                    """
                ))
                if cfg.get("HTTPS_ENABLE", type=bool)
                else Config.Alias(type=str, default="")
            )
        )

    def on_started(self):
        utils.clear_directory(self.get_app_path("conf.d"))
        for container in self.manager.get_installed_containers():
            path = self.get_app_path("temporary", container.name)
            if os.path.isdir(path):
                shutil.copytree(
                    path,
                    self.get_app_path("conf.d", create_parent=True),
                    dirs_exist_ok=True,
                )
        self.manager.create_docker_process(
            "exec", "-it", "nginx",
            "sh", "-c", "killall nginx 1>/dev/null 2>&1"
        ).call()

    def on_removed(self):
        utils.clear_directory(self.get_app_path("temporary"))
        utils.clear_directory(self.get_app_path("conf.d"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@author  : Hu Ji
@file    : deploy.py 
@time    : 2023/05/21
@site    :  
@software: PyCharm 

              ,----------------,              ,---------,
         ,-----------------------,          ,"        ,"|
       ,"                      ,"|        ,"        ,"  |
      +-----------------------+  |      ,"        ,"    |
      |  .-----------------.  |  |     +---------+      |
      |  |                 |  |  |     | -==----'|      |
      |  | $ sudo rm -rf / |  |  |     |         |      |
      |  |                 |  |  |/----|`---=    |      |
      |  |                 |  |  |   ,/|==== ooo |      ;
      |  |                 |  |  |  // |(((( [33]|    ,"
      |  `-----------------'  |," .;'| |((((     |  ,"
      +-----------------------+  ;;  | |         |,"
         /_)______________(_/  //'   | +---------+
    ___________________________/___  `,
   /  oooooooooooooooo  .o.  oooo /,   `,"-----------
  / ==ooooooooooooooo==.o.  ooo= //   ,``--{)B     ,"
 /_==__==========__==_ooo__ooo=_/'   /___________,"
"""
import os
import re
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Any, List, Optional, Callable, TypeVar

import yaml
from jinja2 import Environment, TemplateError

from linktools import utils, Config
from linktools.cli import subcommand, subcommand_argument
from linktools.decorator import cached_property
from linktools.metadata import __missing__
from linktools.rich import choose
from linktools.types import PathType, Error

if TYPE_CHECKING:
    from .manager import ContainerManager
    from linktools._config import ConfigType

    T = TypeVar("T")


class ExposeCategory:

    def __init__(self, name: str, desc: str):
        self.name = name
        self.desc = desc

    def __call__(self, name: str, icon: str, desc: str, url: str):
        return ExposeLink(self, name, icon, desc or name, url)


class ExposeLink:

    def __init__(self, category: ExposeCategory, name: str, icon: str, desc: str, url: str):
        self.category = category
        self.name = name
        self.icon = icon
        self.desc = desc
        self.url = url

    @property
    def is_valid(self) -> bool:
        return not not self.url


class ExposeMixin:
    expose_public = ExposeCategory("public", "Public")
    expose_private = ExposeCategory("private", "Private")
    expose_container = ExposeCategory("container", "Container")
    expose_other = ExposeCategory("other", "Other")

    def load_config_url(self: "BaseContainer", key: str, *path: str):
        url = self.get_config(key, type=str, default=None)
        if url:
            return utils.make_url(url, *path)
        return None

    def load_port_url(self: "BaseContainer", key: str, *path: str, https: bool = True):
        port = self.get_config(key, type=int, default=0)
        if 0 < port < 65535:
            return utils.make_url(f"{'https' if https else 'http'}://{self.manager.host}:{port}", *path)
        return None

    def load_nginx_url(self: "BaseContainer", key: str, *path: str, https: bool = True):
        domain = self.get_config(key, type=str, default=None)
        if domain:
            port = self.get_config("HTTPS_PORT" if https else "HTTP_PORT", type=int)
            return utils.make_url(f"{'https' if https else 'http'}://{domain}:{port}/", *path)
        return None


class NginxMixin:

    def get_nginx_domain(self: "BaseContainer", name: str = None):

        def get_domain(cfg: Config):
            if not self.manager.containers["nginx"].enable:
                return ""
            if not cfg.get("WILDCARD_DOMAIN", type=bool):
                return cfg.get("ROOT_DOMAIN")
            root_domain = cfg.get("ROOT_DOMAIN")
            if root_domain in ("_", "0.0.0.0"):
                return root_domain
            if name is None:
                return f"{self.name}.{root_domain}"
            elif name.strip() == "":
                return root_domain
            return f"{name}.{root_domain}"

        return Config.Lazy(get_domain)

    def write_nginx_conf(self: "BaseContainer", domain: str, template: PathType, name: str = None, https: bool = True):
        nginx = self.manager.containers["nginx"]
        if domain and nginx.enable:
            if not self.get_config("HTTPS_ENABLE", type=bool):
                https = False
            self.render_template(
                nginx.get_path("https.conf" if https else "http.conf"),
                nginx.get_app_path("temporary", self.name, f"{domain}.conf", create_parent=True),
                DOMAIN=domain
            )
            self.render_template(
                template,
                nginx.get_app_path("temporary", self.name, f"{domain}_confs", f"{name or self.name}.conf", create_parent=True),
                DOMAIN=domain
            )


class ContainerError(Error):
    pass


class ContainerTemplateError(ContainerError):
    pass


class AbstractMetaClass(type):

    def __new__(mcs, name, bases, namespace):
        if "__abstract__" not in namespace:
            namespace["__abstract__"] = False
        return super().__new__(mcs, name, bases, namespace)


class BaseContainer(ExposeMixin, NginxMixin, metaclass=AbstractMetaClass):
    __abstract__ = True

    def __init__(self, manager: "ContainerManager", root_path: PathType, name: str = None):
        name = name or self.__module__
        index = name.rfind(".")
        if index >= 0:
            name = name[index + 1:]
        match = re.match(r"^(\d{1,3})-(.*)$", name, re.M | re.I)
        if match:
            self._order = int(match.group(1))
            self._name = match.group(2)
        else:
            self._order = 900
            self._name = name
        self._enable = False
        self.manager = manager
        self.logger = manager.logger
        self.root_path = root_path

    @property
    def name(self) -> str:
        return self._name

    @cached_property
    def description(self) -> str:
        return textwrap.dedent((self.__doc__ or "").strip())

    @property
    def order(self) -> int:
        return self._order

    @property
    def enable(self) -> bool:
        return self._enable

    @enable.setter
    def enable(self, value: bool):
        self._enable = value

    @property
    def dependencies(self) -> [str]:
        return []

    @property
    def configs(self) -> Dict[str, Any]:
        return {}

    @property
    def exposes(self) -> List[ExposeLink]:
        return []

    @cached_property
    def docker_compose(self) -> Optional[Dict[str, Any]]:
        for name in self.manager.docker_compose_names:
            path = self.get_path(name)
            if not os.path.exists(path):
                continue
            data = self.render_template(path)
            data = yaml.safe_load(data)
            if "services" in data and isinstance(data["services"], dict):
                for name, service in data["services"].items():
                    if not isinstance(service, dict):
                        continue
                    service.setdefault("container_name", name)
                    service.setdefault("restart", "unless-stopped")
                    service.setdefault("logging", {
                        "driver": "json-file",
                        "options": {
                            "max-size": "10m",
                        }
                    })
                    if "image" not in service and "build" not in service:
                        path = self.get_docker_file_path()
                        if path and os.path.exists(path):
                            service["build"] = {
                                "context": str(self.get_docker_context_path()),
                                "dockerfile": str(path)
                            }
                    if "env_file" not in service:
                        path = self.get_path(".env")
                        if path and os.path.exists(path):
                            service["env_file"] = [
                                str(path)
                            ]
                return data
        return None

    @cached_property
    def docker_file(self) -> Optional[str]:
        path = self.get_path("Dockerfile")
        if os.path.exists(path):
            return self.render_template(path)
        return None

    @cached_property
    def services(self) -> Dict[str, Dict[str, Any]]:
        services: dict = utils.get_item(self.docker_compose, "services")
        if not services or not isinstance(services, dict):
            return {}
        return services

    @cached_property
    def start_hooks(self) -> List[Callable[[], Any]]:
        return []

    @cached_property
    def stop_hooks(self) -> List[Callable[[], Any]]:
        return []

    def on_init(self):
        pass

    def on_starting(self):
        pass

    def on_started(self):
        pass

    def on_stopping(self):
        pass

    def on_stopped(self):
        pass

    def on_removed(self):
        pass

    @subcommand("shell", help="exec into container using command sh")
    @subcommand_argument("-c", "--command", help="shell command")
    @subcommand_argument("--privileged", help="give extended privileges to the command")
    @subcommand_argument("-u", "--user", help="Username or UID (format: \"<name|uid>[:<group|gid>]\")")
    def on_exec_shell(self, command: str = None, privileged: bool = False, user: str = None):
        service = self.choose_service()

        options = []
        if privileged:
            options.append("--privileged")
        if user:
            options.append("--user")
            options.append(user)

        if not command:
            commands = []
            for shell in ["/bin/zsh", "/bin/fish", "/bin/bash", "/bin/ash", "/bin/sh"]:
                shell_command = [
                    "if" if len(commands) == 0 else "elif", "[", "-x", shell, "]", ";",
                    "then", shell, ";",
                ]
                commands.extend(shell_command)
            commands.extend(["else", "sh", ";"])
            commands.append("fi")
            commands = ("sh", "-c", utils.list2cmdline(commands))
        else:
            commands = utils.cmdline2list(command)

        return self.manager.create_docker_process(
            "exec", "-it", *options, service.get("container_name"), *commands
        ).call()

    @subcommand("logs", help="fetch the logs of container")
    @subcommand_argument("-f", "--follow",
                         help="follow log output")
    @subcommand_argument("-t", "--timestamps",
                         help="show timestamps")
    @subcommand_argument("-n", "--tail", metavar="string",
                         help="number of lines to show from the end of the logs (default \"all\")")
    @subcommand_argument("--since", metavar="string",
                         help="show logs since timestamp (e.g. \"2013-01-02T13:23:37Z\") or relative (e.g. \"42m\" for 42 minutes)")
    @subcommand_argument("--until", metavar="string",
                         help="show logs before a timestamp (e.g. \"2013-01-02T13:23:37Z\") or relative (e.g. \"42m\" for 42 minutes)")
    def on_exec_logs(self, follow: bool = True, tail: str = None, timestamps: bool = True,
                     since: str = None, until: str = None):
        service = self.choose_service()

        options = []
        if follow:
            options.append("--follow")
        if timestamps:
            options.append("--timestamps")
        if tail:
            options.append("--tail")
            options.append(tail)
        if since:
            options.append("--since")
            options.append(since)
        if until:
            options.append("--until")
            options.append(until)
        return self.manager.create_docker_process(
            "logs", *options, service.get("container_name")
        ).call()

    def get_config(self, key: str, type: "ConfigType" = None, default: Any = __missing__) -> "T":
        return self.manager.config.get(key, type=type, default=default)

    def get_path(self, *paths: str) -> Path:
        return utils.join_path(self.root_path, *paths)

    def get_app_path(self, *paths: str, create_parent: bool = False) -> Path:
        path = utils.join_path(self.manager.app_path, self.name, *paths)
        if create_parent:
            path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def get_app_data_path(self, *paths: str, create_parent: bool = False) -> Path:
        path = utils.join_path(self.manager.app_data_path, self.name, *paths)
        if create_parent:
            path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def get_user_data_path(self, *paths: str, create_parent: bool = False) -> Path:
        path = utils.join_path(self.manager.user_data_path, *paths)
        if create_parent:
            path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def get_download_path(self, *paths: str, create_parent: bool = False) -> Path:
        path = utils.join_path(self.manager.download_path, *paths)
        if create_parent:
            path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def get_temp_path(self, *paths: str, create_parent: bool = False) -> Path:
        path = utils.join_path(self.manager.temp_path, "container", self.name, *paths)
        if create_parent:
            path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def choose_service(self) -> Optional[Dict[str, Any]]:
        services = list(self.services.values())
        if len(services) == 0:
            raise ContainerError(f"Not found any service in {self}")
        if len(services) == 1:
            return services[0]
        index = choose(
            "Please choose service",
            choices=[service.get("container_name") for service in services],
            default=0
        )
        return services[index]

    def get_docker_compose_file(self) -> Optional[Path]:
        destination = None
        if self.docker_compose:
            destination = utils.join_path(self.manager.temp_path, "compose", f"{self.name}.yml")
            destination.parent.mkdir(parents=True, exist_ok=True)
            utils.write_file(destination, yaml.dump(self.docker_compose))
        return destination

    def get_docker_file_path(self) -> Optional[Path]:
        destination = None
        if self.docker_file:
            destination = utils.join_path(self.manager.temp_path, "dockerfile", f"{self.name}.Dockerfile")
            destination.parent.mkdir(parents=True, exist_ok=True)
            utils.write_file(destination, self.docker_file)
        return destination

    def get_docker_context_path(self) -> Path:
        return self.get_path()

    def is_depend_on(self, name: str):
        next_items = set(self.dependencies)
        exclude_items = set()
        while next_items:
            if name in next_items:
                return True
            exclude_items.update(next_items)
            current_items = next_items
            next_items = set()
            for next_name in current_items:
                for next_dependency in self.manager.containers[next_name].dependencies:
                    if next_dependency not in exclude_items:
                        next_items.add(next_dependency)
        return False

    def render_template(self, source: PathType, destination: PathType = None, **kwargs: Any):
        config = self.manager.config

        def mkdir(path: PathType) -> str:
            path = config.cast(path, type="path")
            self.start_hooks.append(lambda: os.makedirs(path, mode=0o755, exist_ok=True))
            return path

        def chown(path: PathType, user: str = None) -> str:
            path = config.cast(path, type="path")
            if user:
                uid, gid = utils.get_uid(user), utils.get_gid(user)
                self.start_hooks.append(lambda: self.manager.change_file_owner(path, uid, gid))
            return path

        def chmod(path: PathType, mode: int = 0o755) -> str:
            path = config.cast(path, type="path")
            self.start_hooks.append(lambda: os.chmod(path, mode))
            return path

        context = {
            key: utils.lazy_load(config.get, key)
            for key in config.keys()
        }

        context.update(
            DEBUG=self.manager.debug,

            CONTAINER_PATH=utils.lazy_load(self.get_path),
            APP_PATH=utils.lazy_load(self.get_app_path),
            APP_DATA_PATH=utils.lazy_load(self.get_app_data_path),
            USER_DATA_PATH=utils.lazy_load(self.get_user_data_path),
            DOWNLOAD_PATH=utils.lazy_load(self.get_download_path),

            manager=self.manager,
            container=self,
            config=config,
            user=self.manager.user,
            docker_user=utils.lazy_load(self.get_config, "DOCKER_USER"),

            mkdir=mkdir,
            chown=chown,
            chmod=chmod,
        )

        context.update(kwargs)

        environment = Environment()
        environment.filters.update(
            mkdir=mkdir,
            chown=chown,
            chmod=chmod,
        )

        try:
            self.logger.debug(f"{self} render template {source} to {destination}")
            template = environment.from_string(utils.read_file(source, text=True))
            result = template.render(context)
            if destination:
                utils.write_file(destination, result)
            return result

        except TemplateError as e:
            raise ContainerTemplateError(e)

    def __repr__(self):
        return f"Container<{self.name}>"


class SimpleContainer(BaseContainer):

    def __init__(self, manager: "ContainerManager", root_path: str):
        super().__init__(
            manager,
            root_path,
            name=os.path.basename(root_path)
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@author  : Hu Ji
@file    : repo.py 
@time    : 2024/3/22
@site    : https://github.com/ice-black-tea
@software: PyCharm 

              ,----------------,              ,---------,
         ,-----------------------,          ,"        ,"|
       ,"                      ,"|        ,"        ,"  |
      +-----------------------+  |      ,"        ,"    |
      |  .-----------------.  |  |     +---------+      |
      |  |                 |  |  |     | -==----'|      |
      |  | $ sudo rm -rf / |  |  |     |         |      |
      |  |                 |  |  |/----|`---=    |      |
      |  |                 |  |  |   ,/|==== ooo |      ;
      |  |                 |  |  |  // |(((( [33]|    ,"
      |  `-----------------'  |," .;'| |((((     |  ,"
      +-----------------------+  ;;  | |         |,"
         /_)______________(_/  //'   | +---------+
    ___________________________/___  `,
   /  oooooooooooooooo  .o.  oooo /,   `,"-----------
  / ==ooooooooooooooo==.o.  ooo= //   ,``--{)B     ,"
 /_==__==========__==_ooo__ooo=_/'   /___________,"
"""
import functools
import json
import os
import os.path
import shutil
from typing import TYPE_CHECKING, Dict, Any, List, Union, Callable, Tuple, Set, TypeVar

from filelock import FileLock
from linktools import environ, utils, Config
from linktools.decorator import cached_property
from linktools.types import PathType
from linktools.utils import join_path

from .container import BaseContainer, SimpleContainer, ContainerError
from .repository import Repository

if TYPE_CHECKING:
    from linktools import BaseEnviron

    T = TypeVar("T")


class ContainerManager:

    def __init__(self, environ: "BaseEnviron", name: str = "aio"):  # all_in_one
        self.user = utils.get_user()
        self.uid = utils.get_uid()
        self.gid = utils.get_gid()
        self.system = utils.get_system()
        self.machine = utils.get_machine()

        self.environ = environ
        self.logger = environ.get_logger("container")

        self.config = self.environ.wrap_config(namespace="container", env_prefix="")
        self.config.update_defaults(
            COMPOSE_PROJECT_NAME=name or self.environ.name,
            DOCKER_USER=Config.Prompt(default=os.environ.get("SUDO_USER", self.user).replace(" ", ""), cached=True),
            DOCKER_UID=Config.Lazy(lambda cfg: utils.get_uid(cfg.get("DOCKER_USER", type=str))),
            DOCKER_GID=Config.Lazy(lambda cfg: utils.get_gid(cfg.get("DOCKER_USER", type=str))),
        )

        self.docker_container_name = "container.py"
        self.docker_compose_names = ("compose.yaml", "compose.yml", "docker-compose.yaml", "docker-compose.yml")

        self._file_caches = {}

    @property
    def debug(self) -> bool:
        return os.environ.get("DEBUG", self.environ.debug)

    @cached_property
    def container_type(self) -> str:
        choices = ["docker", "docker-rootless", "podman"] \
            if self.system in ("darwin", "linux") and os.getuid() != 0 \
            else ["docker", "podman"]
        return self.config.get(
            "CONTAINER_TYPE",
            type=str,
            default=Config.Prompt(choices=choices, cached=True),
        )

    @cached_property
    def container_host(self) -> str:
        default = "/var/run/docker.sock"
        host = self.config.get(
            "DOCKER_HOST",
            type=str,
            default=default
        )
        if host:
            left, sep, right = host.partition("://")
            return right or left
        return default

    @cached_property
    def host(self) -> str:
        return self.config.get(
            "HOST",
            type=str,
            default=Config.Prompt(default=utils.get_lan_ip())
        )

    @cached_property
    def app_path(self):
        return self.config.get(
            "DOCKER_APP_PATH",
            type="path",
            default=Config.Prompt(
                default=Config.Lazy(
                    lambda cfg: self.environ.get_data_path("container", "app", create_parent=True)
                ),
                cached=True,
            )
        )

    @cached_property
    def app_data_path(self):
        return self.config.get(
            "DOCKER_APP_DATA_PATH",
            type="path",
            default=Config.Prompt(
                default=Config.Lazy(
                    lambda cfg: self.environ.get_data_path("container", "app_data", create_parent=True)
                ),
                cached=True,
            )
        )

    @cached_property
    def user_data_path(self):
        return self.config.get(
            "DOCKER_USER_DATA_PATH",
            type="path",
            default=Config.Prompt(
                default=Config.Lazy(
                    lambda cfg: self.environ.get_data_path("container", "user_data", create_parent=True)
                ),
                cached=True,
            )
        )

    @cached_property
    def download_path(self):
        return self.config.get(
            "DOCKER_DOWNLOAD_PATH",
            type="path",
            default=Config.Prompt(
                default=Config.Lazy(
                    lambda cfg: self.environ.get_data_path("container", "download", create_parent=True)
                ),
                cached=True,
            )
        )

    @cached_property
    def temp_path(self):
        return self.environ.get_temp_path("container", create_parent=True)

    @cached_property
    def containers(self) -> Dict[str, BaseContainer]:
        result = dict()
        for container in self._load_containers():
            if container.name in result:
                self.logger.debug(f"Container `{container.name}` already exists, overwrite.")
            result[container.name] = container
        with self._config_lock:
            for name in self._load_json_file(self._config_path):
                if name not in result:
                    self.logger.warning(f"Not found installed container `{name}`, skip.")
        return result

    def _load_containers(self) -> List[BaseContainer]:
        containers: List[BaseContainer] = []

        self.logger.debug(f"Load containers from assets")
        asset_path = join_path(os.path.dirname(__file__), "assets", "containers")
        for container in self._walk_containers(asset_path, max_level=1):
            containers.append(container)

        for url, meta in self.get_all_repos().items():
            self.logger.debug(f"Load containers from repository `{url}`")
            repo_path = meta.get("repo_path")
            if not repo_path or not os.path.exists(repo_path) or not os.path.isdir(repo_path):
                self.logger.warning(f"Repository `{url}` not found, skip.")
                continue
            for container in self._walk_containers(repo_path, max_level=2):
                containers.append(container)

        return containers

    def _walk_containers(self, path: PathType, max_level: int):
        if not os.path.isdir(path):
            return
        yield from self._load_container(path)
        if max_level <= 0:
            return
        for name in os.listdir(path):
            yield from self._walk_containers(
                os.path.join(path, name),
                max_level - 1
            )

    def _load_container(self, path: PathType):
        container_path = os.path.join(path, self.docker_container_name)
        if os.path.exists(container_path):
            try:
                name = path.replace(os.sep, ".")
                module = utils.import_module_file(name, container_path)
                for key, value in module.__dict__.items():
                    if isinstance(value, type) and issubclass(value, BaseContainer):
                        if not value.__abstract__:
                            container = value(self, path)
                            self.logger.debug(f"Load container {container.name} in {path}")
                            yield container
                            return
            except Exception as e:
                self.logger.warning(f"Failed to load container from `{path}`: {e}")
                return

        for compose_name in self.docker_compose_names:
            compose_path = os.path.join(path, compose_name)
            if os.path.exists(compose_path):
                container = SimpleContainer(self, path)
                self.logger.debug(f"Load container {container.name} in {path}")
                yield container
                return

    def get_installed_containers(self, resolve: bool = True) -> List[BaseContainer]:
        with self._config_lock:
            containers = self._load_installed_containers()
        if resolve:
            containers = self.resolve_depend_containers(containers)
        return containers

    def resolve_depend_containers(self, containers: List[BaseContainer]) -> List[BaseContainer]:
        order = lambda o: o() if callable(o) else o
        result: Dict[BaseContainer, Union[int, Callable[[], int]]] = dict()
        container_queue: Set[BaseContainer] = set(containers)
        while container_queue:
            container = container_queue.pop()
            result.setdefault(container, container.order)
            for dependency in container.dependencies:
                if dependency not in self.containers:
                    raise ContainerError(f"Dependency container {dependency} not found")
                depend_container = self.containers[dependency]
                if depend_container not in result:
                    result.setdefault(depend_container, depend_container.order)
                    container_queue.add(depend_container)
                if order(result[depend_container]) >= order(result[container]):
                    result[depend_container] = functools.partial(lambda o: order(result[o]) - 1, container)
        return sorted(result, key=lambda o: (order(result[o]), o.order, o.name))

    def prepare_installed_containers(self) -> List[BaseContainer]:
        self.logger.debug(f"Load container type: {self.container_type}")  # 加载容器类型
        containers = self.get_installed_containers(resolve=True)
        if not containers:
            raise ContainerError("No container installed")
        for container in self.containers.values():
            container.enable = container in containers
        for container in reversed(containers):
            self.config.update_defaults(**container.configs)
        for container in containers:
            container.on_init()
        for container in containers:
            if container.docker_file and self.debug:  # 加载每个容器的dockerfile
                self.logger.debug(f"Generate Dockerfile for {container.name}")
            if container.docker_compose and self.debug:  # 加载每个容器的docker-compose.yml
                self.logger.debug(f"Generate docker-compose.yml for {container.name}")
        return containers

    def add_installed_containers(self, *names: str) -> List[BaseContainer]:
        with self._config_lock:
            result = set()
            for name in names:
                container = self.containers.get(name, None)
                if container:
                    result.add(container)
            containers = self._load_installed_containers(reload=True)
            containers.extend(result)
            self._dump_installed_containers(containers)
            return list(result)

    def remove_installed_containers(self, *names: str, force: bool = False) -> List[BaseContainer]:
        with self._config_lock:
            containers = self._load_installed_containers(reload=True)

            result = set()
            remove_names = set(names)
            for name in set(names):
                if name not in self.containers:
                    continue
                for container in containers:
                    if not container.is_depend_on(name):
                        continue
                    if container.name in remove_names:
                        continue
                    if force:
                        remove_names.add(container.name)
                    elif container not in remove_names:
                        raise ContainerError(
                            f"{container} depends on {self.containers[name]}, "
                            f"cannot remove {self.containers[name]}"
                        )

            for name in remove_names:
                container = self.containers.get(name, None)
                if container and container in containers:
                    result.add(container)
                    containers.remove(container)

            self._dump_installed_containers(containers)

            return list(result)

    def _load_installed_containers(self, reload: bool = False) -> List[BaseContainer]:
        result = set()
        for name in self._load_json_file(self._config_path, reload=reload, default=()):
            if name in self.containers:
                result.add(self.containers[name])
        return list(result)

    def _dump_installed_containers(self, containers: List[BaseContainer]) -> None:
        self._dump_json_file(
            self._config_path,
            list(set([container.name for container in containers])),
        )

    def create_process(
            self,
            *args,
            privilege: bool = None,
            **kwargs
    ) -> utils.Process:
        if privilege:
            if self.system in ("darwin", "linux") and self.uid != 0:
                envs = ("http_proxy","https_proxy","all_proxy","no_proxy")
                args = [
                    "sudo",
                    f"--preserve-env={','.join([e.lower() for e in envs] + [e.upper() for e in envs])}",
                    *args
                ]
        return utils.create_process(*args, **kwargs)

    def create_docker_process(
            self,
            *args,
            privilege: bool = None,
            **kwargs
    ) -> utils.Process:
        commands = []
        if self.container_type in ("docker", "docker-rootless"):
            commands.extend(["docker"])
            if privilege is None:
                privilege = self.container_type == "docker"
        elif self.container_type == "podman":
            commands.extend(["podman"])
        else:
            raise ContainerError(f"Invalid container type: {self.container_type}")
        return self.create_process(*commands, *args, privilege=privilege, **kwargs)

    def create_docker_compose_process(
            self,
            containers: List[BaseContainer],
            *args: str,
            privilege: bool = None,
            **kwargs: Any
    ) -> utils.Process:
        options = []
        for container in containers:
            path = container.get_docker_compose_file()
            if path and os.path.exists(path):
                options.extend(["--file", path])
        options.extend(["--project-name", self.config.get("COMPOSE_PROJECT_NAME")])
        return self.create_docker_process("compose", *options, *args, privilege=privilege, **kwargs)

    def change_file_owner(self, path: PathType, uid: int, gid: int, force: bool = False, privilege: bool = False) -> None:
        if not os.path.exists(path):
            self.logger.debug(f"Path not found: {path}")
            return
        if not shutil.which("chown"):
            self.logger.debug("Command `chown` not found")
            return
        stat = os.stat(path)
        if force or stat.st_uid != uid or stat.st_gid != gid:
            try:
                self.create_process(
                    "chown", "-R", f"{uid}:{gid}", path,
                    privilege=privilege or self.uid != stat.st_uid or self.uid != uid
                ).check_call()
            except Exception as e:
                self.logger.warning(f"Failed to change owner of {path}")
                raise e

    def get_all_repos(self) -> Dict[str, Dict[str, str]]:
        with self._repo_lock:
            repos = self._load_json_file(self._repo_config_path)
        return repos

    def add_repo(self, url: str, branch: str = None, force: bool = False):

        with self._repo_lock:
            repos = self._load_json_file(self._repo_config_path, reload=True)

            def ensure_repo_not_exist(key):
                if key not in repos:
                    return
                if not force:
                    raise ContainerError(f"Repository `{key}` already exists.")
                self._remove_repo_file(repos.pop(key))
                self._dump_json_file(self._repo_config_path, repos)

            if url.startswith("http://") or url.startswith("https://") or \
                    url.startswith("ssh://") or url.startswith("git@"):
                ensure_repo_not_exist(url)
                self.logger.info(f"Add git repository: {url}")
                repo_name = utils.guess_file_name(url)
                repo_path = self._choose_repo_path(repo_name)
                Repository.clone_with_progress(url, repo_path, branch)
                repos[url] = dict(type="git", repo_path=repo_path, repo_name=repo_name)

            else:
                path = os.path.abspath(os.path.expanduser(url))
                if not os.path.exists(path) or not os.path.isdir(path):
                    raise ContainerError(f"Invalid local path: {url}")

                ensure_repo_not_exist(path)
                self.logger.info(f"Add local repository: {path}")
                repo_name = utils.guess_file_name(path)
                repo_path = self._choose_repo_path(repo_name)
                os.symlink(path, repo_path, target_is_directory=True)
                repos[path] = dict(type="local", repo_path=repo_path, repo_name=repo_name)

            self._dump_json_file(self._repo_config_path, repos)

    def update_repos(self, force: bool = False):
        for url, meta in self.get_all_repos().items():
            repo_type = meta.get("type", None)
            repo_path = meta.get("repo_path", None)
            if repo_type == "git" and repo_path:
                self.logger.info(f"Update git repository: {url}")
                if not os.path.exists(repo_path):
                    Repository.clone_with_progress(url, repo_path)
                    continue
                repo = Repository(repo_path)
                if repo.is_dirty():
                    if not force:
                        raise ContainerError(f"Repository `{repo_path}` is dirty")
                    self.logger.warning(f"Repository `{repo_path}` is dirty, reset to HEAD")
                    repo.git.reset(hard=True)
                repo.update_with_progress()

    def remove_repo(self, url: str):
        with self._repo_lock:
            repos = self._load_json_file(self._repo_config_path, reload=True)
            if url not in repos:
                raise ContainerError(f"Repository `{url}` not found.")
            self._remove_repo_file(repos.pop(url))
            self._dump_json_file(self._repo_config_path, repos)

    def _choose_repo_path(self, name: str):
        index = 0
        path = os.path.join(self._repo_path, name)
        while os.path.exists(path):
            path = os.path.join(self._repo_path, f"{name}_{index}")
            index += 1
        return path

    def _remove_repo_file(self, repo: Dict[str, str]):
        repo_path = repo.get("repo_path", None)
        if repo_path and os.path.exists(repo_path):
            if os.path.islink(repo_path):
                self.logger.info(f"Remove link {repo_path}")
                os.unlink(repo_path)
            elif os.path.isdir(repo_path):
                self.logger.info(f"Remove directory {repo_path}")
                shutil.rmtree(repo_path, ignore_errors=True)

    @cached_property
    def _config_lock(self):
        return FileLock(self.environ.get_data_path("container", "config", "container.lock", create_parent=True))

    @cached_property
    def _config_path(self):
        return self.environ.get_data_path("container", "config", "containers.yml", create_parent=True)

    @cached_property
    def _repo_lock(self):
        return FileLock(self.environ.get_data_path("container", "repo", "repo.lock", create_parent=True))

    @cached_property
    def _repo_path(self):
        return self.environ.get_data_path("container", "repo", create_parent=True)

    @cached_property
    def _repo_config_path(self):
        return self.environ.get_data_path("container", "repo", "repo.json", create_parent=True)

    def _load_json_file(self, path: PathType, reload: bool = False, default: Any = None) -> Union[Dict, List, Tuple]:
        if reload:
            self._file_caches.pop(path, None)
        elif path in self._file_caches:
            return self._file_caches[path]
        if os.path.exists(path):
            try:
                self._file_caches[path] = json.loads(utils.read_file(path, text=True))
                return self._file_caches[path]
            except Exception as e:
                self.logger.warning(f"Failed to load config file {path}: {e}")
        return default if default is not None else {}

    def _dump_json_file(self, path: PathType, config: Union[Dict, List, Tuple]):
        try:
            self._file_caches.pop(path, None)
            utils.write_file(path, json.dumps(config, indent=2, ensure_ascii=False))
        except Exception as e:
            self.logger.warning(f"Failed to dump config file {path}: {e}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@author  : Hu Ji
@file    : setup.py
@time    : 2018/11/25
@site    :
@software: PyCharm

              ,----------------,              ,---------,
         ,-----------------------,          ,"        ,"|
       ,"                      ,"|        ,"        ,"  |
      +-----------------------+  |      ,"        ,"    |
      |  .-----------------.  |  |     +---------+      |
      |  |                 |  |  |     | -==----'|      |
      |  | $ sudo rm -rf / |  |  |     |         |      |
      |  |                 |  |  |/----|`---=    |      |
      |  |                 |  |  |   ,/|==== ooo |      ;
      |  |                 |  |  |  // |(((( [33]|    ,"
      |  `-----------------'  |," .;'| |((((     |  ,"
      +-----------------------+  ;;  | |         |,"
         /_)______________(_/  //'   | +---------+
    ___________________________/___  `,
   /  oooooooooooooooo  .o.  oooo /,   `,"-----------
  / ==ooooooooooooooo==.o.  ooo= //   ,``--{)B     ,"
 /_==__==========__==_ooo__ooo=_/'   /___________,"
"""
import os

import yaml
from setuptools import setup


def get_root_path(*paths):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), *paths)


def get_src_path(*paths):
    return get_root_path("src", "linktools-cntr", *paths)


if __name__ == '__main__':

    release = os.environ.get("RELEASE", "false").lower() == "true"
    version = os.environ.get("VERSION", "0.0.1.dev0")
    if version.startswith("v"):
        version = version[len("v"):]

    with open(get_root_path("requirements.yml"), "rt", encoding="utf-8") as fd:
        data = yaml.safe_load(fd)
        # install_requires = dependencies + dev-dependencies
        install_requires = data.get("dependencies")
        if not release:
            install_requires.extend(data.get("dev-dependencies"))
        # extras_require = optional-dependencies
        extras_require = data.get("optional-dependencies")
        all_requires = []
        for requires in extras_require.values():
            all_requires.extend(requires)
        extras_require["all"] = all_requires

    setup(
        version=version,
        install_requires=install_requires,
        extras_require=extras_require,
        entry_points={
            "console_scripts": ["ct-cntr = linktools_cntr.__main__:command.main"],
            "linktools_scripts": ["ct-cntr = linktools_cntr.__main__:command.main"],
        },
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@author  : Hu Ji
@file    : repository.py 
@time    : 2024/3/24
@site    : https://github.com/ice-black-tea
@software: PyCharm 

              ,----------------,              ,---------,
         ,-----------------------,          ,"        ,"|
       ,"                      ,"|        ,"        ,"  |
      +-----------------------+  |      ,"        ,"    |
      |  .-----------------.  |  |     +---------+      |
      |  |                 |  |  |     | -==----'|      |
      |  | $ sudo rm -rf / |  |  |     |         |      |
      |  |                 |  |  |/----|`---=    |      |
      |  |                 |  |  |   ,/|==== ooo |      ;
      |  |                 |  |  |  // |(((( [33]|    ,"
      |  `-----------------'  |," .;'| |((((     |  ,"
      +-----------------------+  ;;  | |         |,"
         /_)______________(_/  //'   | +---------+
    ___________________________/___  `,
   /  oooooooooooooooo  .o.  oooo /,   `,"-----------
  / ==ooooooooooooooo==.o.  ooo= //   ,``--{)B     ,"
 /_==__==========__==_ooo__ooo=_/'   /___________,"
"""
from git import Repo, RemoteProgress
from git.util import CallableRemoteProgress

from linktools import utils
from linktools.rich import create_simple_progress


class Repository(Repo):

    def update_with_progress(self):

        with create_simple_progress("message") as progress:

            tasks = {}

            def update_progress(op_code, cur_count, max_count, git_message):
                op_name = self._get_op_name(op_code)
                if not op_name:
                    return

                task_id = tasks.get(op_name)
                if task_id is None:
                    task_id = tasks[op_name] = progress.add_task(
                        op_name,
                        total=None,
                        message=""
                    )

                max_count = utils.int(max_count, default=None)
                cur_count = utils.int(cur_count, default=None)

                message = f"[progress.percentage]" \
                          f"{utils.coalesce(cur_count, '?')}/" \
                          f"{utils.coalesce(max_count, '?')}"
                if git_message:
                    message += f" [progress.data.speed]{git_message}"

                progress.update(
                    task_id,
                    message=message,
                    completed=cur_count,
                    total=max_count
                )

            self.remote().pull(
                progress=CallableRemoteProgress(update_progress),
                allow_unsafe_protocols=True,
                rebase=True,
            )

    @classmethod
    def clone_with_progress(cls, url: str, repo_path: str = None, branch: str = None):

        with create_simple_progress("message") as progress:

            tasks = {}

            options = dict(depth="1")
            if branch:
                options["branch"] = branch

            def update_progress(op_code, cur_count, max_count, git_message):
                op_name = cls._get_op_name(op_code)
                if not op_name:
                    return

                task_id = tasks.get(op_name)
                if task_id is None:
                    task_id = tasks[op_name] = progress.add_task(
                        op_name,
                        total=None,
                        message=""
                    )

                max_count = utils.int(max_count, default=None)
                cur_count = utils.int(cur_count, default=None)

                message = f"[progress.percentage]" \
                          f"{utils.coalesce(cur_count, '?')}/" \
                          f"{utils.coalesce(max_count, '?')}"
                if git_message:
                    message += f" [progress.data.speed]{git_message}"

                progress.update(
                    task_id,
                    message=message,
                    completed=cur_count,
                    total=max_count
                )

            return cls.clone_from(
                url,
                repo_path,
                progress=update_progress,
                allow_unsafe_protocols=True,
                **options
            )

    @classmethod
    def _get_op_name(cls, op_code: int) -> str:
        op_name = ""
        if op_code & RemoteProgress.COUNTING:
            op_name = "Counting objects"
        elif op_code & RemoteProgress.COMPRESSING:
            op_name = "Compressing objects"
        elif op_code & RemoteProgress.WRITING:
            op_name = "Writing objects"
        elif op_code & RemoteProgress.RECEIVING:
            op_name = "Receiving objects"
        elif op_code & RemoteProgress.RESOLVING:
            op_name = "Resolving deltas"
        elif op_code & RemoteProgress.FINDING_SOURCES:
            op_name = "Finding sources"
        elif op_code & RemoteProgress.CHECKING_OUT:
            op_name = "Checking out files"
        return op_name
